import datetime
import json
import os
import tempfile
from functools import cached_property

import numpy as np
import requests
from dateutil.tz import tzutc
from isal import isal_zlib as zlib
from pyproj import CRS, Transformer
from shapely import MultiPolygon, STRtree, Point

from brightsky.db import fetch
from brightsky.settings import settings
from brightsky.utils import USER_AGENT


def _make_dicts(rows):
    return [dict(row) for row in rows]


def weather(
        date, last_date=None, lat=None, lon=None, dwd_station_id=None,
        wmo_station_id=None, source_id=None, max_dist=50000):
    if not last_date:
        last_date = date + datetime.timedelta(days=1)
    if not date.tzinfo:
        date = date.replace(tzinfo=tzutc())
    if not last_date.tzinfo:
        last_date = last_date.replace(tzinfo=tzutc())
    sources_rows = sources(
        lat=lat, lon=lon, dwd_station_id=dwd_station_id, source_id=source_id,
        wmo_station_id=wmo_station_id, max_dist=max_dist,
        observation_types=['historical', 'current', 'forecast'],
        date=date, last_date=last_date)['sources']
    primary_source_ids = {}
    for row in sources_rows:
        primary_source_ids.setdefault(row['observation_type'], row['id'])
    primary_source_ids = list(primary_source_ids.values())
    weather_rows = _weather(date, last_date, primary_source_ids)
    source_ids = [row['id'] for row in sources_rows]
    if len(weather_rows) < int((last_date - date).total_seconds()) // 3600:
        weather_rows = _weather(date, last_date, source_ids)
    _fill_missing_fields(weather_rows, date, last_date, source_ids, True)
    _fill_missing_fields(weather_rows, date, last_date, source_ids, False)
    used_source_ids = {row['source_id'] for row in weather_rows}
    used_source_ids.update(
        source_id
        for row in weather_rows
        for source_id in row.get('fallback_source_ids', {}).values())
    return {
        'weather': weather_rows,
        'sources': [s for s in sources_rows if s['id'] in used_source_ids],
    }


def _weather(date, last_date, source_id, not_null=None, not_null_or=False):
    params = {
        'date': date,
        'last_date': last_date,
        'source_id': source_id,
    }
    where = "timestamp BETWEEN %(date)s AND %(last_date)s"
    order_by = "timestamp"
    if isinstance(source_id, list):
        where += " AND source_id IN %(source_id_tuple)s"
        order_by += ", array_position(%(source_id)s, source_id)"
        params['source_id_tuple'] = tuple(source_id)
    else:
        where += " AND source_id = %(source_id)s"
    if not_null:
        glue = ' OR ' if not_null_or else ' AND '
        constraint = glue.join(f"{x} IS NOT NULL" for x in not_null)
        where += f" AND ({constraint})"
    sql = f"""
        SELECT DISTINCT ON (timestamp) *
        FROM weather
        WHERE {where}
        ORDER BY {order_by}
    """
    return _make_dicts(fetch(sql, params))


IGNORED_MISSING_FIELDS = {
    # Not available in MOSMIX
    'relative_humidity',
    'wind_gust_direction',
    # Not available in recent and historical measurements
    'precipitation_probability',
    'precipitation_probability_6h',
    # Not measured at many stations
    'solar',
}


def _fill_missing_fields(weather_rows, date, last_date, source_ids, partial):
    incomplete_rows = []
    missing_fields = set()
    for row in weather_rows:
        missing_row_fields = set(k for k, v in row.items() if v is None)
        relevant_fields = missing_row_fields.difference(IGNORED_MISSING_FIELDS)
        if relevant_fields:
            incomplete_rows.append((row, missing_row_fields))
            missing_fields.update(relevant_fields)
    if missing_fields:
        min_date = incomplete_rows[0][0]['timestamp']
        max_date = incomplete_rows[-1][0]['timestamp']
        fallback_rows = {
            row['timestamp']: row
            for row in _weather(
                min_date,
                max_date,
                source_ids,
                not_null=missing_fields,
                not_null_or=partial,
            )
        }
        for row, fields in incomplete_rows:
            fallback_row = fallback_rows.get(row['timestamp'])
            if fallback_row:
                row.setdefault('fallback_source_ids', {})
                for f in fields:
                    if fallback_row[f] is None:
                        continue
                    row[f] = fallback_row[f]
                    row['fallback_source_ids'][f] = fallback_row['source_id']


def current_weather(
        lat=None, lon=None, dwd_station_id=None, wmo_station_id=None,
        source_id=None, max_dist=50000, fallback=True):
    sources_rows = sources(
        lat=lat, lon=lon, dwd_station_id=dwd_station_id,
        wmo_station_id=wmo_station_id, source_id=source_id,
        observation_types=['synop'], max_dist=max_dist,
    )['sources']
    source_ids = [row['id'] for row in sources_rows]
    weather = _current_weather(source_ids)
    if not weather:
        raise LookupError(
            "Could not find current weather for your location criteria")
    used_source_ids = [weather['source_id']]
    if fallback:
        missing_fields = [k for k, v in weather.items() if v is None]
        fallback_weather = _current_weather(
            source_ids, not_null=missing_fields)
        if fallback_weather:
            weather.update({k: fallback_weather[k] for k in missing_fields})
            weather['fallback_source_ids'] = {
                field: fallback_weather['source_id']
                for field in missing_fields}
            used_source_ids.append(fallback_weather['source_id'])
    return {
        'weather': weather,
        'sources': [s for s in sources_rows if s['id'] in used_source_ids],
    }


def _current_weather(source_ids, not_null=None):
    params = {
        'source_ids': source_ids,
        'source_ids_tuple': tuple(source_ids),
    }
    where = "source_id IN %(source_ids_tuple)s"
    if not_null:
        where += ''.join(f" AND {element} IS NOT NULL" for element in not_null)
    sql = f"""
        SELECT *
        FROM current_weather
        WHERE {where}
        ORDER BY array_position(%(source_ids)s, source_id)
        LIMIT 1
    """
    rows = _make_dicts(fetch(sql, params))
    if not rows:
        return {}
    return rows[0]


def synop(
        date, last_date=None, dwd_station_id=None, wmo_station_id=None,
        source_id=None):
    if not last_date:
        last_date = date + datetime.timedelta(days=1)
    sources_rows = sources(
        dwd_station_id=dwd_station_id, wmo_station_id=wmo_station_id,
        source_id=source_id, observation_types=['synop'])['sources']
    source_ids = [row['id'] for row in sources_rows]
    sql = """
        SELECT *
        FROM synop
        WHERE
            timestamp BETWEEN %(date)s AND %(last_date)s AND
            source_id IN %(source_ids_tuple)s
        ORDER BY timestamp
        """
    params = {
        'date': date,
        'last_date': last_date,
        'source_ids_tuple': tuple(source_ids),
    }
    return {
        'weather': _make_dicts(fetch(sql, params)),
        'sources': _make_dicts(sources_rows),
    }


def radar(
        date, last_date=None, lat=None, lon=None, distance=200000,
        fmt='compressed', bbox=None):
    extra = {}
    if not last_date:
        last_date = date + datetime.timedelta(hours=2)
    if lat is not None and lon is not None:
        x, y = _transformer.to_xy(lat, lon)
        if not -0.5 <= x <= 1099.5 or not -0.5 <= y <= 1199.5:
            raise LookupError("lat/lon lies outside the radar data range")
        center_x = int(round(x))
        center_y = int(round(y))
        pixels = distance // 1000
        bbox = (
            max(center_y - pixels, 0),
            max(center_x - pixels, 0),
            min(center_y + pixels, 1199),
            min(center_x + pixels, 1099),
        )
        extra['bbox'] = bbox
        extra['latlon_position'] = {
            'x': round(x - bbox[1], 3),
            'y': round(y - bbox[0], 3),
        }
    sql = """
        SELECT *
        FROM radar
        WHERE timestamp BETWEEN %(date)s AND %(last_date)s
        ORDER BY timestamp
        """
    params = {
        'date': date,
        'last_date': last_date,
    }
    rows = fetch(sql, params)
    if fmt == 'plain':
        for row in rows:
            row['precipitation_5'] = _load_radar(row['precipitation_5'], bbox)
    elif fmt == 'bytes':
        for row in rows:
            row['precipitation_5'] = memoryview(
                _load_radar(row['precipitation_5'], bbox),
            )
    elif fmt == 'compressed' and bbox:
        for row in rows:
            row['precipitation_5'] = zlib.compress(
                _load_radar(row['precipitation_5'], bbox),
            )
    elif fmt != 'compressed':
        raise ValueError("Unknown format: '%s'" % fmt)
    return {
        'radar': _make_dicts(rows),
        'geometry': _transformer.bbox_to_geometry(bbox),
        **extra,
    }


def _load_radar(raw, bbox, width=1100, height=1200):
    precip = np.frombuffer(
        zlib.decompress(raw),
        dtype='i2',
    ).reshape((height, width))
    if bbox:
        top, left, bottom, right = bbox
        precip = precip[top:bottom+1, left:right+1]
        # Arrays must be C-contiguous for orjson and zlib
        precip = np.ascontiguousarray(precip).reshape(precip.shape)
    return precip


class RadarCoordinatesTransformer:

    PROJ_STR = (
        "+proj=stere +lat_0=90 +lat_ts=60 +lon_0=10 +a=6378137 "
        "+b=6356752.3142451802 +no_defs +x_0=543196.83521776402 "
        "+y_0=3622588.8619310018"
    )

    @cached_property
    def de1200(self):
        return CRS.from_proj4(self.PROJ_STR)

    @cached_property
    def wgs84_to_de1200(self):
        return Transformer.from_crs(4326, self.de1200)

    @cached_property
    def de1200_to_wgs84(self):
        return Transformer.from_crs(self.de1200, 4326)

    def to_xy(self, lat, lon):
        x, y = self.wgs84_to_de1200.transform(lat, lon)
        return round(x) / 1000, -round(y) / 1000

    def to_latlon(self, x, y):
        lat, lon = self.de1200_to_wgs84.transform(x * 1000, -y * 1000)
        return round(lat, 5), round(lon, 5)

    def to_lonlat(self, x, y):
        return tuple(reversed(self.to_latlon(x, y)))

    def bbox_to_geometry(self, bbox):
        if not bbox:
            bbox = (0, 0, 1199, 1099)
        top, left, bottom, right = bbox
        return {
            'type': 'Polygon',
            'coordinates': [
                self.to_lonlat(left - .5, top - .5),
                self.to_lonlat(left - .5, bottom + .5),
                self.to_lonlat(right + .5, bottom + .5),
                self.to_lonlat(right + .5, top - .5),
            ],
        }


_transformer = RadarCoordinatesTransformer()


def alerts(lat=None, lon=None, warn_cell_id=None):
    if lat is not None and lon is not None:
        meta = _warn_cells.find(lat, lon)
    elif warn_cell_id is not None:
        try:
            meta = _warn_cells.get_meta(warn_cell_id)
        except KeyError:
            raise LookupError(
                "Unknown warn_cell_id, please use commune (Gemeinden), not "
                "district (Landkreis) ids"
            )
    else:
        sql = """
            SELECT *
            FROM alerts
            JOIN (
                SELECT alert_id, array_agg(warn_cell_id) as warn_cell_ids
                FROM alert_cells
                GROUP BY alert_id
            ) cells ON alerts.id = cells.alert_id
            ORDER BY severity DESC
        """
        return {'alerts': _make_dicts(fetch(sql))}
    sql = """
        SELECT *
        FROM alerts
        WHERE id IN (
            SELECT alert_id
            FROM alert_cells
            WHERE warn_cell_id = %(warn_cell_id)s
        )
        ORDER BY severity DESC
        """
    params = {
        'warn_cell_id': meta['warn_cell_id'],
    }
    rows = fetch(sql, params)
    return {
        'alerts': _make_dicts(rows),
        'location': meta,
    }


class WarnCellManager:

    CELLS_CACHE_PATH = os.path.join(tempfile.gettempdir(), 'alert_cells.json')

    @cached_property
    def tree(self):
        self.cell_meta = {}
        self.cell_meta_by_id = {}
        for f in self.get_cell_data()['features']:
            polygons = [
                # shell, holes
                (c[0], c[1:])
                for c in f['geometry']['coordinates']
            ]
            p = MultiPolygon(polygons)
            meta = {
                'warn_cell_id': f['properties']['WARNCELLID'],
                'name': f['properties']['NAME'],
                'name_short': f['properties']['KURZNAME'],
                'district': f['properties']['KREIS'],
                'state': f['properties']['BUNDESLAND'],
                'state_short': f['properties']['BL_KUERZEL'],
            }
            self.cell_meta[p] = meta
            self.cell_meta_by_id[meta['warn_cell_id']] = meta
        return STRtree(list(self.cell_meta.keys()))

    def get_cell_data(self):
        path = self.CELLS_CACHE_PATH
        if not os.path.isfile(path):
            resp = requests.get(
                settings.WARN_CELLS_URL,
                headers={'User-Agent': USER_AGENT},
            )
            with open(path, 'wb') as f:
                f.write(resp.content)
        with open(path) as f:
            return json.load(f)

    def find(self, lat, lon):
        p = Point(lon, lat)
        cell = self.tree.geometries[self.tree.nearest(p)]
        if cell.distance(p) > 0.01:
            raise LookupError("Requested position is not covered by the DWD")
        return self.cell_meta[cell]

    def get_meta(self, warn_cell_id):
        # Make sure cells have been parsed
        self.tree
        return self.cell_meta_by_id[warn_cell_id]


_warn_cells = WarnCellManager()


def sources(
        lat=None, lon=None, dwd_station_id=None, wmo_station_id=None,
        source_id=None, observation_types=None, max_dist=50000,
        ignore_type=False, date=None, last_date=None):
    select = "*"
    order_by = "observation_type"
    params = {
        'lat': lat,
        'lon': lon,
        'max_dist': max_dist,
        'dwd_station_id': dwd_station_id,
        'wmo_station_id': wmo_station_id,
        'source_id': source_id,
        'observation_types': tuple(observation_types or ()),
        'date': date,
        'last_date': last_date,
    }
    if source_id is not None:
        if isinstance(source_id, list):
            where = "id IN %(source_id_tuple)s"
            order_by = (
                "array_position(%(source_id)s, id), observation_type")
            params['source_id_tuple'] = tuple(source_id)
        else:
            where = "id = %(source_id)s"
    elif dwd_station_id is not None:
        if isinstance(dwd_station_id, list):
            where = "dwd_station_id IN %(dwd_station_id_tuple)s"
            order_by = (
                "array_position(%(dwd_station_id)s, dwd_station_id::text), "
                "observation_type")
            params['dwd_station_id_tuple'] = tuple(dwd_station_id)
        else:
            where = "dwd_station_id = %(dwd_station_id)s"
    elif wmo_station_id is not None:
        if isinstance(wmo_station_id, list):
            where = "wmo_station_id IN %(wmo_station_id_tuple)s"
            order_by = (
                "array_position(%(wmo_station_id)s, wmo_station_id::text), "
                "observation_type")
            params['wmo_station_id_tuple'] = tuple(wmo_station_id)
        else:
            where = "wmo_station_id = %(wmo_station_id)s"
    elif (lat is not None and lon is not None):
        distance = """
            earth_distance(
                ll_to_earth(%(lat)s, %(lon)s),
                ll_to_earth(lat, lon)
            )
        """
        select += f", round({distance}) AS distance"
        where = f"""
            earth_box(
                ll_to_earth(%(lat)s, %(lon)s),
                %(max_dist)s
            ) @> ll_to_earth(lat, lon) AND
            {distance} < %(max_dist)s
        """
        if ignore_type:
            order_by = "distance"
        else:
            order_by += ", distance"
    else:
        raise ValueError(
            "Please supply lat/lon or dwd_station_id or wmo_station_id or "
            "source_id")
    if observation_types:
        where += " AND observation_type IN %(observation_types)s"
    if date is not None:
        where += " AND last_record >= %(date)s"
    if last_date is not None:
        where += " AND first_record <= %(last_date)s"
    sql = f"""
        SELECT {select}
        FROM sources
        WHERE {where}
        ORDER BY {order_by}
        """
    rows = fetch(sql, params)
    if not rows:
        raise LookupError("No sources match your criteria")
    return {'sources': _make_dicts(rows)}



import csv
import datetime
import re

import dwdparse.parsers
import numpy as np
from dateutil.tz import tzutc
from isal import isal_zlib as zlib

from brightsky.db import fetch
from brightsky.export import (
    AlertExporter,
    DBExporter,
    RADOLANExporter,
    SYNOPExporter,
)
from brightsky.settings import settings


class BrightSkyMixin:

    PRIORITY = 10
    exporter = DBExporter

    def skip_path(self, path):
        return False


class ObservationsBrightSkyMixin(BrightSkyMixin):

    def skip_path(self, path):
        if (m := re.search(r'_(\d{8})_(\d{8})_hist\.zip$', str(path))):
            end_date = datetime.datetime.strptime(
                m.group(2),
                '%Y%m%d',
            ).replace(tzinfo=tzutc())
            if end_date < settings.MIN_DATE:
                return True
            if settings.MAX_DATE:
                start_date = datetime.datetime.strptime(
                    m.group(1), '%Y%m%d').replace(tzinfo=tzutc())
                if start_date > settings.MAX_DATE:
                    return True
        return False

    def skip_timestamp(self, timestamp):
        if timestamp < settings.MIN_DATE:
            return True
        elif settings.MAX_DATE and timestamp > settings.MAX_DATE:
            return True
        return False


class MOSMIXParser(BrightSkyMixin, dwdparse.parsers.MOSMIXParser):

    PRIORITY = 20


class SYNOPParser(BrightSkyMixin, dwdparse.parsers.SYNOPParser):

    PRIORITY = 30
    exporter = SYNOPExporter


class CurrentObservationsParser(
    BrightSkyMixin,
    dwdparse.parsers.CurrentObservationsParser,
):

    PRIORITY = 30

    def skip_path(self, path):
        return path.endswith(tuple(
            f'{station:_<5}-BEOB.csv'
            for station in settings.IGNORED_CURRENT_OBSERVATIONS_STATIONS
        ))

    def parse(self, path, lat=None, lon=None, height=None, station_name=None):
        if any(x is None for x in (lat, lon, height, station_name)):
            with open(path) as f:
                reader = csv.DictReader(f, delimiter=';')
                wmo_station_id = next(reader)[self.DATE_COLUMN].rstrip('_')
            lat, lon, height, station_name = self._load_location(
                wmo_station_id,
            )
        return super().parse(
            path,
            lat=lat,
            lon=lon,
            height=height,
            station_name=station_name,
        )

    def _load_location(self, wmo_station_id):
        rows = fetch(
            """
            SELECT lat, lon, height, station_name
            FROM sources
            WHERE wmo_station_id = %s
            ORDER BY observation_type DESC, id DESC
            LIMIT 1
            """,
            (wmo_station_id,),
        )
        if not rows:
            raise ValueError(f'Cannot find location for WMO {wmo_station_id}')
        return rows[0]


class CloudCoverObservationsParser(
    ObservationsBrightSkyMixin,
    dwdparse.parsers.CloudCoverObservationsParser,
):
    pass


class DewPointObservationsParser(
    ObservationsBrightSkyMixin,
    dwdparse.parsers.DewPointObservationsParser,
):
    pass


class TemperatureObservationsParser(
    ObservationsBrightSkyMixin,
    dwdparse.parsers.TemperatureObservationsParser,
):
    pass


class PrecipitationObservationsParser(
    ObservationsBrightSkyMixin,
    dwdparse.parsers.PrecipitationObservationsParser,
):
    pass


class SolarRadiationObservationsParser(
    ObservationsBrightSkyMixin,
    dwdparse.parsers.SolarRadiationObservationsParser,
):

    def skip_timestamp(self, timestamp):
        # We aggregate solar radiation from ten-minute data, where the values
        # correspond to radiation for the NEXT ten minutes, i.e. the value
        # tagged 14:30 contains the solar radiation between 14:30 - 14:40 (I
        # have not found a place where this is officially documented, but this
        # interpretation makes the values align with the hourly data from the
        # 'current' sources).
        # This makes solar radiation the only parameter where the 'recent'
        # sources produce a data point for today, which is otherwise only
        # served by the 'current' sources. To avoid excessive fill-up when
        # querying today's weather, we ignore this last data point (but will
        # pick it up on the next day).
        if timestamp.date() == datetime.date.today():
            return True
        return super().skip_timestamp(timestamp)


class VisibilityObservationsParser(
    ObservationsBrightSkyMixin,
    dwdparse.parsers.VisibilityObservationsParser,
):
    pass


class WindObservationsParser(
    ObservationsBrightSkyMixin,
    dwdparse.parsers.WindObservationsParser,
):
    pass


class WindGustsObservationsParser(
    ObservationsBrightSkyMixin,
    dwdparse.parsers.WindGustsObservationsParser,
):
    pass


class SunshineObservationsParser(
    ObservationsBrightSkyMixin,
    dwdparse.parsers.SunshineObservationsParser,
):
    pass


class PressureObservationsParser(
    ObservationsBrightSkyMixin,
    dwdparse.parsers.PressureObservationsParser,
):
    pass


class RADOLANParser(BrightSkyMixin, dwdparse.parsers.RADOLANParser):

    PRIORITY = 30
    exporter = RADOLANExporter

    def process_raw_data(self, raw):
        # XXX: Unlike with the other weather parameters, because of it's large
        #      size, we're storing the radar data in a half-raw state and
        #      performing some final processing during runtime. This brings
        #      down the response time for retrieving one full radar scan
        #      (single timestamp, 1200x1100 pixels) from 1.5 seconds to about
        #      1 ms, mainly because of the reduced data transfer when fetching
        #      the scan from the database. An important caveat of this is that
        #      we are replacing `None` with `0`!
        data = np.array(raw, dtype='i2')
        data[data > 4095] = 0
        data = np.flipud(data.reshape((1200, 1100)))
        return zlib.compress(np.ascontiguousarray(data))


class CAPParser(BrightSkyMixin, dwdparse.parsers.CAPParser):

    PRIORITY = 40
    exporter = AlertExporter


def get_parser(filename):
    parsers = {
        r'DE1200_RV': RADOLANParser,
        r'MOSMIX_(S|L)_LATEST(_240)?\.kmz$': MOSMIXParser,
        r'Z_CAP_C_EDZW_LATEST_.*_COMMUNEUNION_MUL\.zip': CAPParser,
        r'Z__C_EDZW_\d+_.*\.json\.bz2$': SYNOPParser,
        r'\w{5}-BEOB\.csv$': CurrentObservationsParser,
        'stundenwerte_FF_': WindObservationsParser,
        'stundenwerte_N_': CloudCoverObservationsParser,
        'stundenwerte_P0_': PressureObservationsParser,
        'stundenwerte_RR_': PrecipitationObservationsParser,
        'stundenwerte_SD_': SunshineObservationsParser,
        'stundenwerte_TD_': DewPointObservationsParser,
        'stundenwerte_TU_': TemperatureObservationsParser,
        'stundenwerte_VV_': VisibilityObservationsParser,
        '10minutenwerte_extrema_wind_': WindGustsObservationsParser,
        '10minutenwerte_SOLAR_': SolarRadiationObservationsParser,
    }
    for pattern, parser in parsers.items():
        if re.match(pattern, filename):
            return parser



import logging
import os
from contextlib import suppress
from functools import lru_cache

import coloredlogs
import dateutil.parser
from astral import Observer
from astral.sun import daylight

import requests


logger = logging.getLogger(__name__)


USER_AGENT = 'Bright Sky / https://brightsky.dev/'


def configure_logging():
    log_fmt = '%(asctime)s %(name)s %(levelname)s  %(message)s'
    coloredlogs.install(level=logging.DEBUG, fmt=log_fmt)
    # Disable some third-party noise
    logging.getLogger('huey').setLevel(logging.INFO)
    logging.getLogger('urllib3').setLevel(logging.WARNING)


def load_dotenv(path='.env'):
    if not int(os.getenv('BRIGHTSKY_LOAD_DOTENV', 1)):
        return
    with suppress(FileNotFoundError):
        with open(path) as f:
            for line in f:
                if line.strip() and not line.strip().startswith('#'):
                    key, val = line.strip().split('=', 1)
                    os.environ.setdefault(key, val)


def download(url, directory):
    """
    Download a resource from `url` into `directory`, returning its path and
    fingerprint.
    """
    resp = requests.get(url, headers={'User-Agent': USER_AGENT})
    resp.raise_for_status()
    filename = os.path.basename(url)
    path = os.path.join(directory, filename)
    with open(path, 'wb') as f:
        f.write(resp.content)
    fingerprint = {
        'url': url,
        'last_modified': dateutil.parser.parse(resp.headers['Last-Modified']),
        'file_size': int(resp.headers['Content-Length']),
    }
    return path, fingerprint


def parse_date(date_str):
    try:
        return dateutil.parser.isoparse(date_str)
    except ValueError as e:
        # Auto-correct common error of not encoding '+' as '%2b' in URL
        handled_errors = [
            'Inconsistent use of colon separator',
            'Unused components in ISO string',
        ]
        if e.args and e.args[0] in handled_errors and date_str.count(' ') == 1:
            return parse_date(date_str.replace(' ', '+'))
        raise e from None


@lru_cache
def sunrise_sunset(lat, lon, date):
    return daylight(Observer(lat, lon), date)


def daytime(lat, lon, timestamp):
    try:
        sunrise, sunset = sunrise_sunset(lat, lon, timestamp.date())
    except ValueError as e:
        return 'day' if 'above' in e.args[0] else 'night'
    if sunset < sunrise:
        return 'night' if sunset <= timestamp <= sunrise else 'day'
    return 'day' if sunrise <= timestamp <= sunset else 'night'



import datetime

from dateutil.tz import tzutc

from brightsky.settings import Settings

from .utils import environ


def test_settings_loads_environment():
    with environ(BRIGHTSKY_TEST='value'):
        assert Settings().TEST == 'value'


def test_settings_parses_environment_bool():
    assert isinstance(Settings().KEEP_DOWNLOADS, bool)
    with environ(BRIGHTSKY_KEEP_DOWNLOADS='0'):
        assert Settings().KEEP_DOWNLOADS is False
    with environ(BRIGHTSKY_KEEP_DOWNLOADS='1'):
        assert Settings().KEEP_DOWNLOADS is True


def test_settings_parses_environment_date():
    expected = datetime.datetime(2000, 1, 2, tzinfo=tzutc())
    assert isinstance(Settings().MIN_DATE, datetime.datetime)
    with environ(BRIGHTSKY_MIN_DATE='2000-01-02'):
        assert Settings().MIN_DATE == expected
    assert Settings().MAX_DATE is None
    with environ(BRIGHTSKY_MAX_DATE='2000-01-02'):
        assert Settings().MAX_DATE == expected


def test_settings_parses_environment_float():
    assert isinstance(Settings().ICON_RAIN_THRESHOLD, float)
    with environ(BRIGHTSKY_ICON_RAIN_THRESHOLD='0'):
        assert Settings().ICON_RAIN_THRESHOLD == float('0')
    with environ(BRIGHTSKY_ICON_RAIN_THRESHOLD='1.5'):
        assert Settings().ICON_RAIN_THRESHOLD == float('1.5')


def test_settings_parses_environment_list():
    assert isinstance(Settings().CORS_ALLOWED_ORIGINS, list)
    with environ(BRIGHTSKY_CORS_ALLOWED_ORIGINS=''):
        assert Settings().CORS_ALLOWED_ORIGINS == []
    with environ(BRIGHTSKY_CORS_ALLOWED_ORIGINS='a'):
        assert Settings().CORS_ALLOWED_ORIGINS == ['a']
    with environ(BRIGHTSKY_CORS_ALLOWED_ORIGINS='a,b,c'):
        assert Settings().CORS_ALLOWED_ORIGINS == ['a', 'b', 'c']



import json
from multiprocessing import cpu_count

import click
import dwdparse
from falcon.testing import simulate_get
from huey.consumer_options import ConsumerConfig

from brightsky import db, tasks
from brightsky.parsers import get_parser
from brightsky.utils import parse_date
from brightsky.web import app, StandaloneApplication
from brightsky.worker import huey


def dump_records(it):
    for record in it:
        print(json.dumps(record, default=str))


def migrate_callback(ctx, param, value):
    if value:
        db.migrate()


def parse_date_arg(ctx, param, value):
    if not value:
        return
    return parse_date(value)


@click.group()
@click.option(
    '--migrate', help='Migrate database before running command.',
    is_flag=True, is_eager=True, expose_value=False, callback=migrate_callback)
def cli():
    pass


@cli.command()
def migrate():
    """Apply all pending database migrations."""
    db.migrate()


@cli.command()
@click.option('--path', hidden=True)
@click.option('--url', hidden=True)
@click.option(
    '--export/--no-export',
    default=False,
    hidden=True,
)
@click.argument(
    'targets',
    required=False,
    nargs=-1,
    metavar='TARGET [TARGET ...]',
)
def parse(path, url, export, targets):
    """Parse and store observations/forecasts from a URL."""
    # TODO: In v2.2, mark `targets` as required, remove `path`, `url`, and
    #       `export`, and keep only the `elif targets` branch of this function.
    if path or url:
        if path:
            records = dwdparse.parse(path)
        elif url:
            records = dwdparse.parse_url(url)
        if export:
            exporter = get_parser(path or url).exporter()
            exporter.export(list(records))
        else:
            dump_records(records)
        click.echo(
            click.style(
                "WARNING: Parsing DWD files with `python -m brightsky parse` "
                "without exporting to a database is no longer supported and "
                "will be removed in version 2.2. Use `dwdparse` instead.",
                fg='red',
            ),
            err=True,
        )
    elif targets:
        for target in targets:
            tasks.parse(target)
    else:
        raise click.ClickException('Please provide at least one target')


@cli.command()
@click.option(
    '--enqueue/--no-enqueue', default=False,
    help='Enqueue updated files for processing by the worker')
def poll(enqueue):
    """Detect updated files on DWD Open Data Server."""
    files = tasks.poll(enqueue=enqueue)
    if not enqueue:
        dump_records(files)


@cli.command()
def clean():
    """Clean expired forecast and observations from database."""
    tasks.clean()


@cli.command()
@click.option('--workers', default=3, type=int, help='Number of threads')
def work(workers):
    """Start brightsky worker."""
    huey.flush()
    config = ConsumerConfig(worker_type='thread', workers=workers)
    config.validate()
    consumer = huey.create_consumer(**config.values)
    consumer.run()


@cli.command()
@click.option('--bind', default='127.0.0.1:5000', help='Bind address')
@click.option(
    '--reload/--no-reload', default=False,
    help='Reload server on source code changes')
def serve(bind, reload):
    """Start brightsky API webserver."""
    StandaloneApplication(
        'brightsky.web:app',
        bind=bind,
        workers=1 if reload else 2*cpu_count()+1,
        reload=reload
    ).run()


@cli.command(context_settings={'ignore_unknown_options': True})
@click.argument('endpoint')
@click.argument('parameters', nargs=-1, type=click.UNPROCESSED)
def query(endpoint, parameters):
    """Query API and print JSON response.

    Parameters must be supplied as --name value or --name=value. See
    https://brightsky.dev/docs/ for the available endpoints and arguments.

    \b
    Examples:
    python -m brightsky query weather --lat 52 --lon 7.6 --date 2018-08-13
    python -m brightsky query current_weather --lat=52 --lon=7.6
    """
    if not app._router.find(f'/{endpoint}'):
        raise click.UsageError(f"Unknown endpoint '{endpoint}'")
    resp = simulate_get(app, f'/{endpoint}', params=_parse_params(parameters))
    print(json.dumps(resp.json))


def _parse_params(parameters):
    # I'm sure there's a function in click or argparse somewhere that does this
    # but I can't find it
    usage = "Supply API parameters as --name value or --name=value"
    params = {}
    param_name = None
    for param in parameters:
        if param_name is None:
            if not param.startswith('--'):
                raise click.UsageError(usage)
            param = param[2:]
            if '=' in param:
                name, value = param.split('=', 1)
                params[name] = value
            else:
                param_name = param
        else:
            params[param_name] = param
            param_name = None
    if param_name is not None:
        raise click.UsageError(usage)
    return params
